@@ -31,6 +31,19 @@ def raise_exception(e):
     raise e
 
 
+
+# --- global helper functions ---
+
+
+
+# dictionary of global helper functions to be added to the Jinja2 environment;
+# as each function is defined, they are added to this and then it iterated
+# through, to extend the environment later
+
+global_helpers = {}
+
+
+
 def warn_helper(msg):
     """Helper function to print a warning inside a Jinja2 template.  The
     supplied message is printed to stderr but execution is not stopped.
@@ -50,6 +63,10 @@ def warn_helper(msg):
     return ""
 
 
+global_helpers["warn"] = warn_helper
+
+
+
 def raise_helper(msg):
     """Helper function to raise an exception inside a Jinja2 template.  The
     generic 'Exception' class is raised in Python.
@@ -62,6 +79,10 @@ def raise_helper(msg):
     raise Exception(msg)
 
 
+global_helpers["raise"] = raise_helper
+
+
+
 def assert_helper(condition, msg):
     """Helper function for assertions inside a Jinja2 template.  If the
     supplied condition is False, an exception is raised with the given
@@ -85,6 +106,10 @@ def assert_helper(condition, msg):
     return ""
 
 
+global_helpers["assert"] = assert_helper
+
+
+
 def re_match_helper(pattern, string):
     """Helper function for matching strings against regular expressions
     in a Jinja2 template: a boolean is returned, indicating whether the
@@ -102,6 +127,10 @@ def re_match_helper(pattern, string):
     return bool(re.match(pattern, string))
 
 
+global_helpers["re_match"] = re_match_helper
+
+
+
 def re_match_group_helper(pattern, string, *groups):
     """Helper function for matching strings against regular expressions
     and extract selected groups (marked by parentheses) of text in a
@@ -135,6 +164,10 @@ def re_match_group_helper(pattern, string, *groups):
     return r.group(*groups)
 
 
+global_helpers["re_match_group"] = re_match_group_helper
+
+
+
 def re_match_groups_helper(pattern, string):
     """Helper function for matching strings against regular expressions
     and extract groups (marked by parentheses) of text in a Jinja2
@@ -164,6 +197,71 @@ def re_match_groups_helper(pattern, string):
     return r.groups()
 
 
+global_helpers["re_match_groups"] = re_match_groups_helper
+
+
+
+def build_dict_helper(s):
+    """Helper function to build a dictionary from a string: the lines
+    of the string are parsed in the form 'key:value' and used to
+    populate a dictionary that is returned.
+
+    Blank lines (including ones consisting entirely of spaces) are
+    skipped, as well as leading and trailing spaces in the key and value.
+
+    The function is useful to get more complex data structures back
+    from Jinja2 macros (which can only return text).
+    """
+
+    d = {}
+    
+    for i in s.split("\n"):
+        # skip blank lines (including ones just composed of spaces)
+        if not i.strip():
+            continue
+
+        try:
+            k, v = i.split(":", 1)
+        except ValueError:
+            raise(ValueError("failed parsing 'key:value' from line: %s" % i))
+
+        # store the key:value in the dictionary, trimming spaces
+        d[k.strip()] = v.strip()
+
+    return d
+
+
+global_helpers["build_dict"] = build_dict_helper
+
+
+
+def build_list_helper(s):
+    """Helper function to build a list from the lines in a multiline
+    string: each line is added as an item to the list.
+
+    Blank lines (including ones consiting entirely of spaces) are
+    skipped, as well as leading and trailing spaces on the remaining
+    lines.
+
+    The function is useful to get more complex data structures back
+    from Jinja2 macros (which can only return text).
+    """
+
+    l = []
+
+    for i in s.split("\n"):
+        i = i.strip()
+        
+        # only add to the list, if the line is not blank
+        if i:
+            l.append(i)
+
+    return l
+
+
+global_helpers["build_list"] = build_list_helper
+
+
 
 # --- command line arguments ---
 
@@ -219,6 +317,13 @@ parser.add_argument(
     help="generate the configuration but do not output it - useful to test "
          "generation succeeds")
 
+parser.add_argument(
+    "-E", "--list-env",
+    action="store_true",
+    help="list filter and global helper functions and stop (this will "
+          "include the standard Jinja2, as well as the netaddr module and "
+          "custom ones)")
+
 parser.add_argument(
     "-I", "--dump-inventory",
     action="store_true",
@@ -270,6 +375,7 @@ include_dirname = os.path.join(args.config_dirname, args.include_dirname)
 inventory_dirname = os.path.join(args.config_dirname, args.inventory_dirname)
 output_filename = args.output_filename
 no_output = args.no_output
+list_env = args.list_env
 dump_inventory = args.dump_inventory
 dump_device = args.dump_device
 quiet = args.quiet
@@ -468,17 +574,8 @@ env = jinja2.Environment(
 # add in the special warn(), raise() and assert() functions, as well as
 # some other functions we need, into the Jinja2 environment
 
-env.globals["warn"] = warn_helper
-env.globals["raise"] = raise_helper
-env.globals["assert"] = assert_helper
-env.globals["re_match"] = re_match_helper
-env.globals["re_match_group"] = re_match_group_helper
-env.globals["re_match_groups"] = re_match_groups_helper
-
-env.globals["deepcopy"] = deepcopy
-
-env.globals["deepmerge"] = deepmerge
-env.globals["deepremoveitems"] = deepremoveitems
+for global_helper in global_helpers:
+        env.globals[global_helper] = global_helpers[global_helper]
 
 
 # add in the netaddr library functions as additional filters
@@ -489,6 +586,23 @@ for filter_name, filter_func in (
     env.filters[filter_name] = filter_func
 
 
+# if the 'list environment' option is specified, print out the filter and
+# global helper functions and stop
+
+if list_env:
+    print("Filters:")
+    for filter in sorted(env.filters):
+        print("  " + filter)
+
+    print()
+
+    print("Global helpers:")
+    for global_helper in sorted(env.globals):
+        print("  " + global_helper)
+
+    exit(0)
+
+
 
 def profiles_merge(source_description, device, included_profiles=set()):
     """This function recursively merges profiles into the supplied
